// 東方老桜夢 用 トークデータコンパイラ

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <ctype.h>
#include <memory.h>
#include <windows.h>

#include "dimiConfigLoader.h"

#define STRMAX  1024
#define DICTMAX 1024
#define SCMAX   (16*1024)
#define SFMAX   64
#define BMPMAX  (64*1024)

#define CH_TY_NONE         0
#define CH_TY_SINGLE_0     1
#define CH_TY_SINGLE_1     2
#define CH_TY_SINGLE_2     3
#define CH_TY_VOICED_0     4
#define CH_TY_VOICED_1     5
#define CH_TY_SEMI_VOICED  6
#define CH_TY_DOUBLE       7
#define CH_TY_MARK_0       8
#define CH_TY_MARK_1       9

#define SC_QUIT    0x01
#define SC_WAIT    0x02
#define SC_SPACE   0x03
#define SC_RETURN  0x04
#define SC_FUNC    0x0c
#define SC_BGM     0x10
#define SC_BALLOON 0x20
#define SC_PLAYER  0x28
#define SC_ENEMY   0x80

#define CMD_QUIT   0xff
#define CMD_WAIT   0xfe
#define CMD_RETURN 0xfd
#define CMD_DISP   0xf0
#define CMD_BGM    0xef
#define CMD_FUNC   0xee

typedef struct {
	int type;
	int ch;
	int xpos;
	int ypos;
	int base;
	int assign;
} DICT;

DICT g_dict_tbl[DICTMAX];
int  g_dict_num;

int  g_sc_data[SCMAX];
int  g_sc_num;
int *g_sf_ptr[SFMAX];
int  g_sf_size[SFMAX];
int  g_sf_num;

char g_bmp[BMPMAX];
int  g_bmp_width;
int  g_bmp_height;
int  g_bmp_pitch;

char g_tmp1[STRMAX];
char g_tmp2[STRMAX];

char *Utf8ToLocal(const char *src, char *dst, int size) {
	wchar_t buf[STRMAX];
	int len;

	len = MultiByteToWideChar(CP_UTF8, 0, src, strlen(src)+1, buf, STRMAX);
	WideCharToMultiByte(CP_ACP, 0, buf, len, dst, size, NULL, NULL);

	return dst;
}

char *WideToUtf8(int ch, char *buf) {

	if (ch >= 0x04000000) {
		buf[0] = (uint8_t)(((ch>>30) & 0x01) | 0xfc);
		buf[1] = (uint8_t)(((ch>>24) & 0x3f) | 0x80);
		buf[2] = (uint8_t)(((ch>>18) & 0x3f) | 0x80);
		buf[3] = (uint8_t)(((ch>>12) & 0x3f) | 0x80);
		buf[4] = (uint8_t)(((ch>> 6) & 0x3f) | 0x80);
		buf[5] = (uint8_t)(((ch    ) & 0x3f) | 0x80);
		buf[6] = '\0';
	} else if (ch >= 0x200000) {
		buf[0] = (uint8_t)(((ch>>24) & 0x03) | 0xf8);
		buf[1] = (uint8_t)(((ch>>18) & 0x3f) | 0x80);
		buf[2] = (uint8_t)(((ch>>12) & 0x3f) | 0x80);
		buf[3] = (uint8_t)(((ch>> 6) & 0x3f) | 0x80);
		buf[4] = (uint8_t)(((ch    ) & 0x3f) | 0x80);
		buf[5] = '\0';
	} else if (ch >= 0x010000) {
		buf[0] = (uint8_t)(((ch>>18) & 0x07) | 0xf0);
		buf[1] = (uint8_t)(((ch>>12) & 0x3f) | 0x80);
		buf[2] = (uint8_t)(((ch>> 6) & 0x3f) | 0x80);
		buf[3] = (uint8_t)(((ch    ) & 0x3f) | 0x80);
		buf[4] = '\0';
	} else if (ch >= 0x0800) {
		buf[0] = (uint8_t)(((ch>>12) & 0x0f) | 0xe0);
		buf[1] = (uint8_t)(((ch>> 6) & 0x3f) | 0x80);
		buf[2] = (uint8_t)(((ch    ) & 0x3f) | 0x80);
		buf[3] = '\0';
	} else if (ch >= 0x80) {
		buf[0] = (uint8_t)(((ch>> 6) & 0x1f) | 0xc0);
		buf[1] = (uint8_t)(((ch    ) & 0x3f) | 0x80);
		buf[2] = '\0';
	} else {
		buf[0] = (uint8_t)ch;
		buf[1] = '\0';
	}

	return buf;
}

int ReadLine(FILE *fp, int *buf, size_t size) {
	int *p;
	int c;
	int w;

	w = 0;
	p = buf;
	while (p < buf+size) {
		c = fgetc(fp);
		if (c < 0) break;

		if (c >= 0xfc) {
			if (w > 0) *(p++) = w;
			w = c & 0x01;
		} else if (c >= 0xf8) {
			if (w > 0) *(p++) = w;
			w = c & 0x03;
		} else if (c >= 0xf0) {
			if (w > 0) *(p++) = w;
			w = c & 0x07;
		} else if (c >= 0xe0) {
			if (w > 0) *(p++) = w;
			w = c & 0x0f;
		} else if (c >= 0xc0) {
			if (w > 0) *(p++) = w;
			w = c & 0x1f;
		} else if (c >= 0x80) {
			w <<= 6;
			w |= c & 0x3f;
		} else if (c == 0x0a) {
			if (w > 0) *(p++) = w;
			w = 0;
			break;
		} else {
			if (w > 0) *(p++) = w;
			*(p++) = c & 0x7f;
			w = 0;
		}
	}

	if (w > 0) *(p++) = w;
	*p = 0;

	return ((c < 0) && ((p - buf) == 0)) ? -1 : (int)(p - buf);
}

int *ParseInt(int *ptr, int *num) {
	int minus = 0;
	int valid = 0;
	int *p;
	int n = 0;

	p = ptr;
	while ((*p < 0x0100) && isspace(*p)) p++;

	if (*p == '-') {
		minus = 1;
		p++;
	} else if (*p == '+') {
		p++;
	}
	if (*p == '$') {
		p++;
		while ((*p < 0x0100) && isxdigit(*p)) {
			if (islower(*p)) {
				n = (n * 16) + (*p - 'a' + 10);
			} else if (isupper(*p)) {
				n = (n * 16) + (*p - 'A' + 10);
			} else {
				n = (n * 16) + (*p - '0');
			}
			valid = 1;
			p++;
		}
	} else {
		while ((*p < 0x0100) && isdigit(*p)) {
			n = (n * 10) + (*p - '0');
			valid = 1;
			p++;
		}
	}
	if (valid == 0) return ptr;

	if (minus) n = -n;
	*num = n;

	return p;
}

int SearchDictionaryByChar(int ch) {
	int i;

	for (i = 0; i < g_dict_num; i++) {
		if (g_dict_tbl[i].ch == ch) return i;
	}

	return -1;
}


int ReadDictionary(char *name) {
	FILE *fp;
	int buf[STRMAX];
	int line;
	int *p, *q;
	int n;
	int r;

	fp = fopen(name, "r");
	if (fp == NULL) {
		printf("file open error: %s\n", name);
		return 1;
	}

	line = 0;
	g_dict_num = 0;
	while (g_dict_num < DICTMAX) {
		r = ReadLine(fp, buf, STRMAX);
		if (r < 0) break;
		line++;

		p = buf;
		while ((*p < 0x0100) && isspace(*p)) p++;
		if (*p == '\0') {
			;
		} else if (*p == '#') {
			;
		} else if (*p == 's') {
			p++;
			if (*p == '0') {
				g_dict_tbl[g_dict_num].type = CH_TY_SINGLE_0;
			} else if (*p == '1') {
				g_dict_tbl[g_dict_num].type = CH_TY_SINGLE_1;
			} else if (*p == '2') {
				g_dict_tbl[g_dict_num].type = CH_TY_SINGLE_2;
			} else {
				printf("%s %d: unknown type s%s\n", name, line, *p);
				continue;
			}
			p++;

			while ((*p < 0x0100) && isspace(*p)) p++;
			if (*p < 0x0100) {
				printf("%s %d: invalid char %c\n", name, line, *p);
				continue;
			}
			r = SearchDictionaryByChar(*p);
			if (r >= 0) {
				printf("%s %d: already defined %s\n", name, line,
					Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
				continue;
			}
			g_dict_tbl[g_dict_num].ch = *(p++);

			q = ParseInt(p, &n);
			if (q <= p) {
				printf("%s %d: not integer (xpos) %s\n", name, line,
					Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
				continue;
			}
			p = q;
			g_dict_tbl[g_dict_num].xpos = n;

			q = ParseInt(p, &n);
			if (q <= p) {
				printf("%s %d: not integer (ypos) %s\n", name, line,
					Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
				continue;
			}
			p = q;
			g_dict_tbl[g_dict_num].ypos = n;

			g_dict_tbl[g_dict_num].assign = 0;
			g_dict_num++;
		} else if (*p == 'b') {
			p++;
			if (*p == '0') {
				g_dict_tbl[g_dict_num].type = CH_TY_VOICED_0;
			} else if (*p == '1') {
				g_dict_tbl[g_dict_num].type = CH_TY_VOICED_1;
			} else {
				printf("%s %d: unknown type b%c\n", name, line, *p);
				continue;
			}
			p++;

			while ((*p < 0x0100) && isspace(*p)) p++;
			if (*p < 0x0100) {
				printf("%s %d: invalid char %c\n", name, line, *p);
				continue;
			}
			r = SearchDictionaryByChar(*p);
			if (r >= 0) {
				printf("%s %d: already defined %s\n", name, line,
					Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
				continue;
			}
			g_dict_tbl[g_dict_num].ch = *(p++);

			while ((*p < 0x0100) && isspace(*p)) p++;
			if (*p < 0x0100) {
				printf("%s %d: invalid char %c\n", name, line, *p);
				continue;
			}
			r = SearchDictionaryByChar(*p);
			if (r < 0) {
				printf("%s %d: undefined base char %s\n", name, line,
					Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
				continue;
			}
			g_dict_tbl[g_dict_num].base = *(p++);

			g_dict_tbl[g_dict_num].assign = 0;
			g_dict_num++;
		} else if (*p == 'p') {
			p++;
			g_dict_tbl[g_dict_num].type = CH_TY_SEMI_VOICED;

			while ((*p < 0x0100) && isspace(*p)) p++;
			if (*p < 0x0100) {
				printf("%s %d: invalid char %c\n", name, line, *p);
				continue;
			}
			r = SearchDictionaryByChar(*p);
			if (r >= 0) {
				printf("%s %d: already defined %s\n", name, line,
					Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
				continue;
			}
			g_dict_tbl[g_dict_num].ch = *(p++);

			while ((*p < 0x0100) && isspace(*p)) p++;
			if (*p < 0x0100) {
				printf("%s %d: invalid char %c\n", name, line, *p);
				continue;
			}
			r = SearchDictionaryByChar(*p);
			if (r < 0) {
				printf("%s %d: undefined base char %s\n", name, line,
					Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
				continue;
			}
			g_dict_tbl[g_dict_num].base = *(p++);

			g_dict_tbl[g_dict_num].assign = 0;
			g_dict_num++;
		} else if (*p == 'w') {
			p++;
			g_dict_tbl[g_dict_num].type = CH_TY_DOUBLE;

			while ((*p < 0x0100) && isspace(*p)) p++;
			if (*p < 0x0100) {
				printf("%s %d: invalid char %c\n", name, line, *p);
				continue;
			}
			r = SearchDictionaryByChar(*p);
			if (r >= 0) {
				printf("%s %d: already defined %s\n", name, line,
					Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
				continue;
			}
			g_dict_tbl[g_dict_num].ch = *(p++);

			q = ParseInt(p, &n);
			if (q <= p) {
				printf("%s %d: not integer (xpos) %s\n", name, line,
					Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
				continue;
			}
			p = q;
			g_dict_tbl[g_dict_num].xpos = n;

			q = ParseInt(p, &n);
			if (q <= p) {
				printf("%s %d: not integer (ypos) %s\n", name, line,
					Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
				continue;
			}
			p = q;
			g_dict_tbl[g_dict_num].ypos = n;

			g_dict_tbl[g_dict_num].assign = 0;
			g_dict_num++;
		} else if (*p == 'm') {
			p++;
			if (*p == '0') {
				g_dict_tbl[g_dict_num].type = CH_TY_MARK_0;
			} else if (*p == '1') {
				g_dict_tbl[g_dict_num].type = CH_TY_MARK_1;
			} else {
				printf("%s %d: unknown type m%c\n", name, line, *p);
				continue;
			}
			p++;

			q = ParseInt(p, &n);
			if (q <= p) {
				printf("%s %d: not integer (xpos) %s\n", name, line,
					Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
				continue;
			}
			p = q;
			g_dict_tbl[g_dict_num].xpos = n;

			q = ParseInt(p, &n);
			if (q <= p) {
				printf("%s %d: not integer (ypos) %s\n", name, line,
					Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
				continue;
			}
			p = q;
			g_dict_tbl[g_dict_num].ypos = n;

			g_dict_tbl[g_dict_num].assign = 0;
			g_dict_num++;
		} else {
			printf("%s %d: unknown type %c\n", name, line, *p);
		}
	}

	fclose(fp);

	return 0;
}

int ReadScript(char *name) {
	FILE *fp;
	int buf[STRMAX];
	int line;
	int *p, *q;
	int n;
	int r;

	fp = fopen(name, "r");
	if (fp == NULL) {
		printf("file open error: %s\n", name);
		return 1;
	}

	line = 0;
	while (g_sc_num < SCMAX) {
		r = ReadLine(fp, buf, STRMAX);
		if (r < 0) break;
		line++;

		p = buf;
		while (*p) {
			while ((*p < 0x0100) && isspace(*p)) p++;
			if (*p >= 0x0100) {
				if (*p == 0x3000) {
					g_sc_data[g_sc_num++] = SC_SPACE;
				} else {
					r = SearchDictionaryByChar(*p);
					if (r >= 0) {
						g_dict_tbl[r].assign = -1;
						g_sc_data[g_sc_num++] = *p;
					} else {
						printf("%s %d: not found in dictionary %s\n", name, line,
							Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
					}
				}
				p++;
			} else if (*p == '<') {
				p++;
				if (*p == 'q') {
					g_sc_data[g_sc_num++] = SC_QUIT;
				} else if (*p == 'w') {
					g_sc_data[g_sc_num++] = SC_WAIT;
				} else if (*p == 'r') {
					g_sc_data[g_sc_num++] = SC_RETURN;
				} else if (*p == 'p') {
					p++;
					if (('0' <= *p) && (*p <= '6')) {
						g_sc_data[g_sc_num++] = SC_PLAYER + (*p - '0');
					} else {
						printf("%s %d: invalid tag p%s\n", name, line,
							Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
					}
				} else if (*p == 'e') {
					p++;
					if (isdigit(*p)) {
						n = 0;
						while (isdigit(*p)) {
							n = n * 10 + (*p - '0');
							p++;
						}
						r = n % 10;
						n = (n / 10) & 0x0f;
						if (n > 0) {
							n = (n << 2) + ((r-1) & 3);
						}
						g_sc_data[g_sc_num++] = SC_ENEMY + n;
					} else {
						printf("%s %d: invalid tag e%s\n", name, line,
							Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
					}
				} else if (*p == 'b') {
					p++;
					if (('0' <= *p) && (*p <= '3')) {
						g_sc_data[g_sc_num++] = SC_BALLOON + (*p - '0');
					} else {
						printf("%s %d: invalid tag b%s\n", name, line,
							Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
					}
				} else if (*p == 'm') {
					p++;
					if (isdigit(*p)) {
						n = 0;
						while (isdigit(*p)) {
							n = n * 10 + (*p - '0');
							p++;
						}
						g_sc_data[g_sc_num++] = SC_BGM + (n & 15);
					} else {
						printf("%s %d: invalid tag m%s\n", name, line,
							Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
					}
				} else if (*p == 'f') {
					p++;
					if (isdigit(*p)) {
						n = 0;
						while (isdigit(*p)) {
							n = n * 10 + (*p - '0');
							p++;
						}
						g_sc_data[g_sc_num++] = SC_FUNC + (n & 3);
					} else {
						printf("%s %d: invalid tag f%s\n", name, line,
							Utf8ToLocal(WideToUtf8(*p, g_tmp1), g_tmp2, STRMAX));
					}
				}
				while (*p && (*p != '>')) p++;
				if (*p) p++;
			} else {
				p++;
			}
		}
	}

	fclose(fp);

	return 0;
}

int AssignCode(int chmax) {
	int pos;
	int i, r;

	// 濁音 半濁音 に対応する清音の割り当て指定
	for (i = 0; i < g_dict_num; i++) {
		if (g_dict_tbl[i].assign &&
		    ((g_dict_tbl[i].type == CH_TY_VOICED_0) ||
		     (g_dict_tbl[i].type == CH_TY_VOICED_1) ||
		     (g_dict_tbl[i].type == CH_TY_SEMI_VOICED))) {
			r = SearchDictionaryByChar(g_dict_tbl[i].base);
			if (r >= 0) {
				g_dict_tbl[r].assign = -1;
			}
		}
	}

	pos = 2;

	// は行の文字の割り当て
	for (i = 0; i < g_dict_num; i++) {
		if (g_dict_tbl[i].assign && g_dict_tbl[i].type == CH_TY_SINGLE_0) {
			printf("%3d:%s\n", pos, Utf8ToLocal(WideToUtf8(g_dict_tbl[i].ch, g_tmp1), g_tmp2, STRMAX));
			g_dict_tbl[i].assign = pos;
			pos++;
		}
	}

	// か行さ行た行の割り当て
	for (i = 0; i < g_dict_num; i++) {
		if (g_dict_tbl[i].assign && g_dict_tbl[i].type == CH_TY_SINGLE_1) {
			printf("%3d:%s\n", pos, Utf8ToLocal(WideToUtf8(g_dict_tbl[i].ch, g_tmp1), g_tmp2, STRMAX));
			g_dict_tbl[i].assign = pos;
			pos++;
		}
	}

	// 清音の割り当て
	for (i = 0; i < g_dict_num; i++) {
		if (g_dict_tbl[i].assign && g_dict_tbl[i].type == CH_TY_SINGLE_2) {
			printf("%3d:%s\n", pos, Utf8ToLocal(WideToUtf8(g_dict_tbl[i].ch, g_tmp1), g_tmp2, STRMAX));
			g_dict_tbl[i].assign = pos;
			pos++;
		}
	}

	// 漢字の割り当て
	if (pos & 1) pos++;
	for (i = 0; i < g_dict_num; i++) {
		if (g_dict_tbl[i].assign && g_dict_tbl[i].type == CH_TY_DOUBLE) {
			printf("%3d:%s\n", pos, Utf8ToLocal(WideToUtf8(g_dict_tbl[i].ch, g_tmp1), g_tmp2, STRMAX));
			g_dict_tbl[i].assign = pos;
			pos += 2;
		}
	}

	if (pos > chmax) {
		printf("assign overflow\n");
		return 1;
	}

	return 0;
}

int WriteData(char *name, char *label, int *data, int size) {
	FILE *fp;
	int buf[4];
	int *p, *limit;
	int display;
	int num;
	int col;
	int i, r;

	fp = fopen(name, "w");
	if (fp == NULL) {
		printf("file open error: %s\n", name);
		return 1;
	}
	fprintf(fp, "%s:\n", label);

	display = 0;

	col = 0;
	num = 0;
	p = data;
	limit = data + size;
	while (p < limit) {
		if ((*p < SC_BALLOON) || (*p >= SC_ENEMY+119)) {
			if (display & 0x1000) {
				display &= 0x07ff;
				buf[num++] = CMD_DISP + ((display>>8) & 0x07);
				buf[num++] = display & 0xff;
			}
		}

		if (*p >= 0x0100) {
			r = SearchDictionaryByChar(*p);
			switch (g_dict_tbl[r].type) {
			case CH_TY_SINGLE_0:
			case CH_TY_SINGLE_1:
			case CH_TY_SINGLE_2:
				buf[num++] = 0x40 + g_dict_tbl[r].assign-2;
				break;
			case CH_TY_VOICED_0:
			case CH_TY_VOICED_1:
				r = SearchDictionaryByChar(g_dict_tbl[r].base);
				buf[num++] = 0xb8 + g_dict_tbl[r].assign-2;
				break;
			case CH_TY_SEMI_VOICED:
				r = SearchDictionaryByChar(g_dict_tbl[r].base);
				buf[num++] = 0xe0 + g_dict_tbl[r].assign-2;
				break;
			case CH_TY_DOUBLE:
				buf[num++] = 0x00 + g_dict_tbl[r].assign / 2;
				break;
			}
		} else if (*p == SC_QUIT) {
			buf[num++] = CMD_QUIT;
		} else if (*p == SC_WAIT) {
			buf[num++] = CMD_WAIT;
		} else if (*p == SC_SPACE) {
			buf[num++] = 0x00;
		} else if (*p == SC_RETURN) {
			buf[num++] = CMD_RETURN;
		} else if ((*p >= SC_FUNC) && (*p <= SC_FUNC+3)) {
			buf[num++] = CMD_FUNC;
			buf[num++] = *p & 0x03;
		} else if ((*p >= SC_BGM) && (*p <= SC_BGM+15)) {
			buf[num++] = CMD_BGM;
			buf[num++] = *p & 0x0f;
		} else if ((*p >= SC_BALLOON) && (*p <= SC_BALLOON+3)) {
			display &= 0xff3f;
			display |= 0x1000 | ((*p & 0x03) << 6);
		} else if ((*p >= SC_PLAYER) && (*p <= SC_PLAYER+6)) {
			display &= 0xf8ff;
			display |= 0x1000 | ((*p & 0x07) << 8);
		} else if ((*p >= SC_ENEMY) && (*p <= SC_ENEMY+119)) {
			display &= 0xffc0;
			display |= 0x1000 | (*p & 0x3f);
		}
		p++;

		for (i = 0; i < num; i++) {
			if (col == 0) {
				fprintf(fp, "\t.db\t$%02x", buf[i]);
				col++;
			} else if (col < 7) {
				fprintf(fp, ", $%02x", buf[i]);
				col++;
			} else {
				fprintf(fp, ", $%02x\n", buf[i]);
				col = 0;
			}
		}
		num = 0;
	}

	if (col > 0) {
		fprintf(fp, "\n");
	}

	fclose(fp);

	return 0;
}

int ReadBMP(char *name) {
	FILE *fp;
	char head[64];
	int pitch, offset;
	int h;

	fp = fopen(name, "rb");
	if (fp == NULL) {
		printf("file open error: %s\n", name);
		return 1;
	}

	fread(head, 54, sizeof(char), fp);
	g_bmp_width   = (head[18] & 255);
	g_bmp_width  |= (head[19] & 255) <<  8;
	g_bmp_width  |= (head[20] & 255) << 16;
	g_bmp_width  |= (head[21] & 255) << 24;
	g_bmp_height  = (head[22] & 255);
	g_bmp_height |= (head[23] & 255) <<  8;
	g_bmp_height |= (head[24] & 255) << 16;
	g_bmp_height |= (head[25] & 255) << 24;
	pitch = (g_bmp_width+3) & -4;

	offset = 54 + 1024;
	offset += pitch * g_bmp_height;
	for (h = 0; h < g_bmp_height; h++) {
		offset -= pitch;
		fseek(fp, offset, SEEK_SET);
		fread(g_bmp+pitch*h, pitch, sizeof(char), fp);
	}

	fclose(fp);

	return 0;
}

int WriteCHR(char *name, int chmax) {
	FILE *fp;
	char buf[16*128];
	int pitch;
	int pos, num, x, y;
	int data1, data2;
	int i, j, w, h;

	fp = fopen(name, "wb");
	if (fp == NULL) {
		printf("file open error: %s\n", name);
		return 1;
	}

	pitch = (g_bmp_width+3) & -4;

	memset(buf, 0xff, sizeof(buf));
	for (i = 0; i < g_dict_num; i++) {
		num = 0;
		x = g_dict_tbl[i].xpos;
		y = g_dict_tbl[i].ypos;
		if (g_dict_tbl[i].type == CH_TY_MARK_0) {
			pos = 0;
			num = 1;
		} else if (g_dict_tbl[i].type == CH_TY_MARK_1) {
			pos = 1;
			num = 1;
		} else if (g_dict_tbl[i].assign) {
			switch (g_dict_tbl[i].type) {
			case CH_TY_SINGLE_0:
			case CH_TY_SINGLE_1:
			case CH_TY_SINGLE_2:
				pos = g_dict_tbl[i].assign;
				num = 1;
				break;
			case CH_TY_DOUBLE:
				pos = g_dict_tbl[i].assign;
				num = 2;
				break;
			}
		}
		for (j = 0; j < num; j++) {
			for (h = 0; h < 8; h++) {
				data1 = 0;
				data2 = 0;
				for (w = 0; w < 8; w++) {
					data1 |= ((g_bmp[(y+h)*pitch+(x+w)]   ) & 1) << (7-w);
					data2 |= ((g_bmp[(y+h)*pitch+(x+w)]>>1) & 1) << (7-w);
				}
				buf[pos*16+h  ] = data1;
				buf[pos*16+h+8] = data2;
			}
			pos++;
			y += 8;
		}
	}

	fwrite(buf, 16*chmax, sizeof(char), fp);

	fclose(fp);

	return 0;
}

int main(int argc, char *argv[]) {
	void *conf;
	char *name;
	char *label;
	int chmax;
	int i, r;

	if (argc < 2) {
		printf("Usage: %s <param file>\n", argv[0]);
		return 1;
	}

	g_sc_num = 0;
	g_sf_num = 0;

	r = DimiConf_Open(&conf, argv[1]);
	if (r) {
		printf("file open error: %s\n", argv[1]);
		return 1;
	}

	name = DimiConf_GetString(conf, "", "dictionary", NULL);
	if (name == NULL) {
		printf("dictionary file name is not define.\n");
		goto err;
	}
	r = ReadDictionary(name);
	if (r) {
		goto err;
	}

	chmax = strtol(DimiConf_GetString(conf, "", "chr_max", (char *)"0"), NULL, 0);
	if ((chmax <= 0) || (chmax > 128)) {
		chmax = 128;
	}

	while (g_sf_num < SFMAX) {
		sprintf(g_tmp1, "in_script%d", g_sf_num);
		name = DimiConf_GetString(conf, "", g_tmp1, NULL);
		if (name == NULL) {
			break;
		}
		g_sf_ptr[g_sf_num] = g_sc_data + g_sc_num;

		r = ReadScript(name);

		g_sf_size[g_sf_num] = (g_sc_data + g_sc_num) - g_sf_ptr[g_sf_num];
		g_sf_num++;
	}

	r = AssignCode(chmax);
	if (r) {
		goto err;
	}

	for (i = 0; i < g_sf_num; i++) {
		sprintf(g_tmp1, "out_data%d", i);
		name = DimiConf_GetString(conf, "", g_tmp1, NULL);
		if (name == NULL) {
			break;
		}
		sprintf(g_tmp1, "out_label%d", i);
		label = DimiConf_GetString(conf, "", g_tmp1, NULL);
		if (label == NULL) {
			break;
		}

		r = WriteData(name, label, g_sf_ptr[i], g_sf_size[i]);
	}

	name = DimiConf_GetString(conf, "", "in_bmp", NULL);
	if (name == NULL) {
		printf("bmp file name is not define.\n");
		goto err;
	}
	r = ReadBMP(name);
	if (r) {
		goto err;
	}

	name = DimiConf_GetString(conf, "", "out_chr", NULL);
	if (name == NULL) {
		printf("chr file name is not define.\n");
		goto err;
	}
	r = WriteCHR(name, chmax);
	if (r) {
		goto err;
	}

	DimiConf_Close(conf);

	return 0;

err:
	DimiConf_Close(conf);

	return 1;
}
